@@ -5,6 +5,7 @@
 #include <time.h>
 #include <string.h>
 #include <stdint.h>
+#include <inttypes.h>
 #include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -99,6 +100,7 @@ void print_DbType(const long int *type) {
             case DBTYPE_PFO_ADVOK: printf("PFO_ADVOK\n"); break;
             case DBTYPE_PFO_DANPOR: printf("PFO_DAPOR\n"); break;
             case DBTYPE_PFO_INSSPR: printf("PFO_INSSPR\n"); break;
+            case DBTYPE_PFO_AUDITOR: printf("PFO_AUDITOR\n"); break;
             case DBTYPE_PO: printf("PO\n"); break;
             case DBTYPE_PO_ZAK: printf("PO_ZAK\n"); break;
             case DBTYPE_PO_REQ: printf("PO_REQ\n"); break;
@@ -106,6 +108,9 @@ void print_DbType(const long int *type) {
             case DBTYPE_OVM_NOTAR: printf("OVM_NOTAR\n"); break;
             case DBTYPE_OVM_EXEKUT: printf("OVM_EXEKUT\n"); break;
             case DBTYPE_OVM_REQ: printf("OVM_REQ\n"); break;
+            case DBTYPE_OVM_FO: printf("OVM_FO\n"); break;
+            case DBTYPE_OVM_PFO: printf("OVM_PFO\n"); break;
+            case DBTYPE_OVM_PO: printf("OVM_PO\n"); break;
             default: printf("<unknown type %ld>\n", *type);
         }
 }
@@ -119,6 +124,10 @@ void print_UserType(const long int *type) {
             case USERTYPE_ENTRUSTED: printf("ENTRUSTED\n"); break;
             case USERTYPE_ADMINISTRATOR: printf("ADMINISTRATOR\n"); break;
             case USERTYPE_OFFICIAL: printf("OFFICIAL\n"); break;
+            case USERTYPE_OFFICIAL_CERT: printf("OFFICIAL_CERT\n"); break;
+            case USERTYPE_LIQUIDATOR: printf("LIQUIDATOR\n"); break;
+            case USERTYPE_RECEIVER: printf("RECEIVER\n"); break;
+            case USERTYPE_GUARDIAN: printf("GUARDIAN\n"); break;
             default: printf("<unknown type %ld>\n", *type);
         }
 }
@@ -133,6 +142,10 @@ void print_sender_type(const isds_sender_type *type) {
             case SENDERTYPE_ADMINISTRATOR: printf("ADMINISTRATOR\n"); break;
             case SENDERTYPE_OFFICIAL: printf("OFFICIAL\n"); break;
             case SENDERTYPE_VIRTUAL: printf("VIRTUAL\n"); break;
+            case SENDERTYPE_OFFICIAL_CERT: printf("OFFICIAL_CERT\n"); break;
+            case SENDERTYPE_LIQUIDATOR: printf("LIQUIDATOR\n"); break;
+            case SENDERTYPE_RECEIVER: printf("RECEIVER\n"); break;
+            case SENDERTYPE_GUARDIAN: printf("GUARDIAN\n"); break;
             default: printf("<unknown type %u>\n", *type);
         }
 }
@@ -379,16 +392,21 @@ void print_DbUserInfo(const struct isds_DbUserInfo *info) {
 void print_timeval(const struct timeval *time) {
     struct tm broken;
     char buffer[128];
+    time_t seconds_as_time_t;
 
     if (!time) {
         printf("NULL\n");
         return;
     }
-    
-    if (!localtime_r(&(time->tv_sec), &broken)) goto error;
+
+    /* MinGW32 GCC 4.8+ uses 64-bit time_t but time->tv_sec is defined as
+     * 32-bit long in Microsoft API. Convert value to the type expected by
+     * gmtime_r(). */
+    seconds_as_time_t = time->tv_sec;
+    if (!localtime_r(&seconds_as_time_t, &broken)) goto error;
     if (!strftime(buffer, sizeof(buffer)/sizeof(char), "%c", &broken))
         goto error;
-    printf("%s, %jd us\n", buffer, (intmax_t)time->tv_usec);
+    printf("%s, %" PRIdMAX " us\n", buffer, (intmax_t)time->tv_usec);
     return;
 
 error:
@@ -724,7 +742,7 @@ int mmap_file(const char *file, int *fd, void **buffer, size_t *length) {
         return -1;
     }
     if (file_info.st_size < 0) {
-        fprintf(stderr, "File `%s' has negative size: %jd\n", file,
+        fprintf(stderr, "File `%s' has negative size: %" PRIdMAX "\n", file,
                 (intmax_t) file_info.st_size);
         close(*fd);
         return -1;
