@@ -26,10 +26,12 @@ extern "C" {
 extern const char isds_locator[];       /* Without client certificate auth. */
 extern const char isds_cert_locator[];  /* With client certificate auth. */
 extern const char isds_otp_locator[];   /* With OTP authentication */
+extern const char isds_mep_locator[];   /* Mobile key application. */
 /* Base URL of testing ISDS instance */
 extern const char isds_testing_locator[];       /* Without client certificate */
 extern const char isds_cert_testing_locator[];  /* With client certificate */
 extern const char isds_otp_testing_locator[];   /* With OTP authentication */
+extern const char isds_mep_testing_locator[];   /* Mobile key application. */
 
 
 struct isds_ctx;    /* Context for specific ISDS box */
@@ -142,10 +144,10 @@ typedef enum {
     OTP_TIME                /* Time-based OTP method */
 } isds_otp_method;
 
-/* One-time passwed authentication resolution */
+/* One-time password authentication resolution */
 typedef enum {
-    OTP_RESOLUTION_SUCCESS = 0,         /* Authentication succeded */
-    OTP_RESOLUTION_UNKNOWN,             /* Status is unkown */
+    OTP_RESOLUTION_SUCCESS = 0,         /* Authentication succeeded */
+    OTP_RESOLUTION_UNKNOWN,             /* Status is unknown */
     OTP_RESOLUTION_BAD_AUTHENTICATION,  /* Bad log-in, retry */
     OTP_RESOLUTION_ACCESS_BLOCKED,      /* Access blocked for 60 minutes
                                            (brute force attack detected) */
@@ -176,21 +178,52 @@ struct isds_otp {
                                        authentication attempt. */
 };
 
+/* Mobile key authentication resolution. */
+typedef enum {
+    MEP_RESOLUTION_SUCCESS = 0,    /* Authentication succeeded */
+    MEP_RESOLUTION_UNKNOWN,        /* Status is unknown */
+    MEP_RESOLUTION_UNRECOGNISED,   /* Authentication request not recognised. */
+    MEP_RESOLUTION_ACK_REQUESTED,  /* Waiting for acknowledgement. */
+    MEP_RESOLUTION_ACK,            /* Acknowledged. */
+    MEP_RESOLUTION_ACK_EXPIRED     /* Acknowledgement request expired. */
+} isds_mep_resolution;
+
+/* Mobile key context to authenticate client */
+struct isds_mep {
+    /* Input members. */
+    char *app_name;                 /* Client application name. This name is
+                                       displayed in the mobile key authentication
+                                       application to provide the user brief
+                                       information about which application is
+                                       requiring access to the data box. */
+    /* Intermediate members. */
+    char *intermediate_uri;         /* Intermediate authentication URI. */
+    /* Output members. */
+    isds_mep_resolution resolution; /* Fine-grade resolution of mobile key
+                                       authentication attempt. */
+};
+
 /* Box type */
 typedef enum {
+    DBTYPE_OVM_MAIN = -1,       /* Special value for
+                                   isds_find_box_by_fulltext(). */
     DBTYPE_SYSTEM = 0,          /* This is special sender value for messages
                                    sent by ISDS. */
     DBTYPE_OVM = 10,
-    DBTYPE_OVM_NOTAR = 11,
-    DBTYPE_OVM_EXEKUT = 12,
+    DBTYPE_OVM_NOTAR = 11,      /* This type has been replaced with OVM_PFO. */
+    DBTYPE_OVM_EXEKUT = 12,     /* This type has been replaced with OVM_PFO. */
     DBTYPE_OVM_REQ = 13,
+    DBTYPE_OVM_FO = 14,
+    DBTYPE_OVM_PFO = 15,
+    DBTYPE_OVM_PO = 16,
     DBTYPE_PO = 20,
-    DBTYPE_PO_ZAK = 21,
+    DBTYPE_PO_ZAK = 21,         /* This type has been replaced with PO. */
     DBTYPE_PO_REQ = 22,
     DBTYPE_PFO = 30,
     DBTYPE_PFO_ADVOK = 31,
     DBTYPE_PFO_DANPOR = 32,
     DBTYPE_PFO_INSSPR = 33,
+    DBTYPE_PFO_AUDITOR = 34,
     DBTYPE_FO = 40
 } isds_DbType;
 
@@ -337,7 +370,9 @@ typedef enum {
     USERTYPE_ADMINISTRATOR,         /* User to manage ENTRUSTED_USERs */
     USERTYPE_OFFICIAL,              /* ??? */
     USERTYPE_OFFICIAL_CERT,         /* ??? */
-    USERTYPE_LIQUIDATOR             /* Company liquidator */
+    USERTYPE_LIQUIDATOR,            /* Company liquidator */
+    USERTYPE_RECEIVER,              /* Company receiver */
+    USERTYPE_GUARDIAN               /* Legal guardian */
 } isds_UserType;
 
 /* Data about user.
@@ -378,8 +413,10 @@ typedef enum {
     EVENT_PRIMARY_LOGIN,            /* Primary user has logged in */
     EVENT_ENTRUSTED_LOGIN,          /* Entrusted user with capability to read
                                        has logged in */
-    EVENT_SYSCERT_LOGIN             /* Application authenticated by `system'
+    EVENT_SYSCERT_LOGIN,            /* Application authenticated by `system'
                                        certificate has logged in */
+    EVENT_UNDELIVERED_AV_CHECK      /* An attachment didn't pass the antivirus
+                                       check, message has not been delivered */
 } isds_event_type;
 
 /* Message event
@@ -622,7 +659,7 @@ struct isds_message {
                                        document of type FILEMETATYPE_MAIN and
                                        can contain any number of other type
                                        documents. Total size of documents
-                                       must not exceed 10 MB. */
+                                       must not exceed 20 MB. */
 };
 
 /* Message copy recipient and assigned message ID */
@@ -723,7 +760,7 @@ struct isds_credit_event_storage_set {
                                        change; Optional. */
 };
 
-/* Event about change of credit for sending commerical services */
+/* Event about change of credit for sending commercial services */
 struct isds_credit_event {
     /* Common fields */
     struct timeval *time;           /* When the credit was changed. */
@@ -733,7 +770,7 @@ struct isds_credit_event {
                                        The unit is 1/100 CZK. */
     isds_credit_event_type type;    /* Type of the event */
 
-    /* Datails specific for the type */
+    /* Details specific for the type */
     union {
         struct isds_credit_event_charged charged;
                                                 /* ISDS_CREDIT_CHARGED */
@@ -771,8 +808,10 @@ typedef enum {
     SENDERTYPE_OFFICIAL,            /* ISDS; sender of system message */
     SENDERTYPE_VIRTUAL,             /* An application (e.g. document
                                        information system) */
-    SENDERTYPE_OFFICIAL_CERT,        /* ???; Non-normative */
-    SENDERTYPE_LIQUIDATOR           /* Liquidator of the company; Non-normative */
+    SENDERTYPE_OFFICIAL_CERT,       /* ???; Non-normative */
+    SENDERTYPE_LIQUIDATOR,          /* Liquidator of the company; Non-normative */
+    SENDERTYPE_RECEIVER,            /* Receiver of the company */
+    SENDERTYPE_GUARDIAN             /* Legal guardian */
 } isds_sender_type;
 
 /* Digital delivery of credentials */
@@ -832,6 +871,15 @@ struct isds_fulltext_result {
                                    into this box */
 };
 
+/* A box state valid in the time range */
+struct isds_box_state_period {
+    struct timeval from;        /* Time range beginning */
+    struct timeval to;          /* Time range end */
+    long int dbState;           /* Box state; 1 <=> active box, otherwise
+                                   inaccessible; use isds_DbState enum to
+                                   identify some states. */
+};
+
 /* Initialize ISDS library.
  * Global function, must be called before other functions.
  * If it fails you can not use ISDS library and must call isds_cleanup() to
@@ -877,7 +925,7 @@ void isds_set_logging(const unsigned int facilities,
  * @facility is log message class
  * @level is log message severity
  * @message is string with zero byte terminator. This can be any arbitrary
- * chunk of a sentence with or without new line, a sentence can be splitted
+ * chunk of a sentence with or without new line, a sentence can be split
  * into more messages. However it should not happen. If you discover message
  * without new line, report it as a bug.
  * @length is size of @message string in bytes excluding trailing zero
@@ -968,6 +1016,8 @@ isds_error isds_set_opt(struct isds_ctx *context, const isds_option option,
  * authentication.
  * @otp selects one-time password authentication method to use, defines OTP
  * code (if known) and returns fine grade resolution of OTP procedure.
+ * @mk selects mobile key authentication method to use, specifies the
+ * communication code.
  * @return:
  *  IE_SUCCESS if authentication succeeds
  *  IE_NOT_LOGGED_IN if authentication fails. If OTP authentication has been
@@ -983,6 +1033,27 @@ isds_error isds_login(struct isds_ctx *context, const char *url,
         const struct isds_pki_credentials *pki_credentials,
         struct isds_otp *otp);
 
+/* Connect and log into ISDS server using the MEP login method.
+ * All arguments are copied, you don't have to keep them after successful
+ * return.
+ * @url is base address of ISDS web service. Pass extern isds_mep_locator to use
+ * the production ISDS environment (pass extern isds_mep_testing_locator to
+ * access the testing environment). Passing null causes the production
+ * environment locator to be used.
+ * @username is the username of ISDS user or box ID
+ * @code is the communication code. The code is generated when enabling
+ * the mobile key authentication and can be found in the web-based portal
+ * of the data-box service.
+ * @return:
+ *  IE_SUCCESS if authentication succeeds
+ *  IE_NOT_LOGGED_IN if authentication fails
+ *  IE_PARTIAL_SUCCESS if MEP authentication has been requested, fine-grade
+ *  resolution is returned via @mep->resolution, keep arguments unchanged and
+ *  repeat the function call as long as IE_PARTIAL_SUCCESS is being returned;
+ *  or other appropriate error. */
+isds_error isds_login_mep(struct isds_ctx *context, const char *url,
+        const char *username, const char *code, struct isds_mep *mep);
+
 /* Log out from ISDS server and close connection. */
 isds_error isds_logout(struct isds_ctx *context);
 
@@ -990,7 +1061,12 @@ isds_error isds_logout(struct isds_ctx *context);
  * Send dummy request to ISDS and expect dummy response. */
 isds_error isds_ping(struct isds_ctx *context);
 
-/* Get data about logged in user and his box. */
+/* Get data about logged in user and his box.
+ * @context is session context
+ * @db_owner_info is reallocated box owner description. It will be freed on
+ * error.
+ * @return error code from lower layer, context message will be set up
+ * appropriately. */
 isds_error isds_GetOwnerInfoFromLogin(struct isds_ctx *context,
         struct isds_DbOwnerInfo **db_owner_info);
 
@@ -1195,10 +1271,11 @@ isds_error isds_delete_user(struct isds_ctx *context,
  * @context is session context
  * @list_identifier is UTF-8 encoded string identifying boxes of interrest.
  * System recognizes following values currently: ALL (all boxes), UPG
- * (effectively OVM boxes), OVM (OVM gross type boxes), OPN (boxes allowing
- * receiving commercial messages). This argument is a string because
- * specification states new values can appear in the future. Not all list
- * types are available to all users.
+ * (effectively OVM boxes), POA (active boxes allowing receiving commercial
+ * messages), OVM (OVM gross type boxes), OPN (boxes allowing receiving
+ * commercial messages). This argument is a string because specification
+ * states new values can appear in the future. Not all list types are
+ * available to all users.
  * @buffer is automatically reallocated memory to store the list of boxes. The
  * list is zipped CSV file.
  * @buffer_length is size of @buffer data in bytes.
@@ -1228,7 +1305,8 @@ isds_error isds_FindDataBox(struct isds_ctx *context,
  * @target selects box attributes to search for @query words. Pass NULL if you
  * don't care.
  * @box_type restricts searching to given box type. Value DBTYPE_SYSTEM means
- * to search in all box types. Pass NULL to let server to use default value
+ * to search in all box types. Value DBTYPE_OVM_MAIN means to search in
+ * non-subsudiary OVM box types. Pass NULL to let server to use default value
  * which is DBTYPE_SYSTEM.
  * @page_size defines count of boxes to constitute a response page. It counts
  * from zero. Pass NULL to let server to use a default value (50 now).
@@ -1284,6 +1362,28 @@ isds_error isds_find_box_by_fulltext(struct isds_ctx *context,
 isds_error isds_CheckDataBox(struct isds_ctx *context, const char *box_id,
         long int *box_status);
 
+/* Get history of box state changes.
+ * @context is ISDS session context.
+ * @box_id is UTF-8 encoded sender box identifier as zero terminated string.
+ * @from_time is first second of history to return in @history. Server ignores
+ * subseconds. NULL means time of creating the box.
+ * @to_time is last second of history to return in @history. Server ignores
+ * subseconds. It's valid to have the @from_time equaled to the @to_time. The
+ * interval is closed from both ends. NULL means now.
+ * @history outputs auto-reallocated list of pointers to struct
+ * isds_box_state_period. Each item describes a continues time when the box
+ * was in one state. The state is 1 for accessible box. Otherwise the box
+ * is inaccessible (priviledged users will get exact box state as enumerated
+ * in isds_DbState, other users 0).
+ * @return:
+ *  IE_SUCCESS if the history has been obtained correctly,
+ *  or other appropriate error. Please note that server allows to retrieve
+ *  the history only to some users. */
+isds_error isds_get_box_state_history(struct isds_ctx *context,
+        const char *box_id,
+        const struct timeval *from_time, const struct timeval *to_time,
+        struct isds_list **history);
+
 /* Get list of permissions to send commercial messages.
  * @context is ISDS session context.
  * @box_id is UTF-8 encoded sender box identifier as zero terminated string
@@ -1833,6 +1933,9 @@ void isds_credentials_delivery_free(
 void isds_fulltext_result_free(
         struct isds_fulltext_result **result);
 
+/* Deallocate struct isds_box_state_period recursively and NULL it */
+void isds_box_state_period_free(struct isds_box_state_period **period);
+
 /* Copy structure isds_PersonName recursively */
 struct isds_PersonName *isds_PersonName_duplicate(
         const struct isds_PersonName *src);
@@ -1849,6 +1952,10 @@ struct isds_DbOwnerInfo *isds_DbOwnerInfo_duplicate(
 struct isds_DbUserInfo *isds_DbUserInfo_duplicate(
         const struct isds_DbUserInfo *src);
 
+/* Copy structure isds_box_state_period recursively */
+struct isds_box_state_period *isds_box_state_period_duplicate(
+        const struct isds_box_state_period *src);
+
 #ifdef __cplusplus  /* For C++ linker sake */
 }
 #endif
