@@ -32,7 +32,7 @@ static int test_dbtype(const isds_DbType type, const xmlChar *name) {
         FAIL_TEST("conversion from isds_DbType to string failed");
 
     if (xmlStrcmp(name, string))
-            FAIL_TEST("Wrong to string conversion result");
+        FAIL_TEST("Wrong to string conversion result");
 
     err = string2isds_DbType(string, &new_type);
     if (err)
@@ -52,6 +52,9 @@ int main(void) {
         DBTYPE_OVM_NOTAR,
         DBTYPE_OVM_EXEKUT,
         DBTYPE_OVM_REQ,
+        DBTYPE_OVM_FO,
+        DBTYPE_OVM_PFO,
+        DBTYPE_OVM_PO,
         DBTYPE_PO,
         DBTYPE_PO_ZAK,
         DBTYPE_PO_REQ,
@@ -59,6 +62,7 @@ int main(void) {
         DBTYPE_PFO_ADVOK,
         DBTYPE_PFO_DANPOR,
         DBTYPE_PFO_INSSPR,
+        DBTYPE_PFO_AUDITOR,
         DBTYPE_FO
     };
 
@@ -67,6 +71,9 @@ int main(void) {
         BAD_CAST "OVM_NOTAR",
         BAD_CAST "OVM_EXEKUT",
         BAD_CAST "OVM_REQ",
+        BAD_CAST "OVM_FO",
+        BAD_CAST "OVM_PFO",
+        BAD_CAST "OVM_PO",
         BAD_CAST "PO",
         BAD_CAST "PO_ZAK",
         BAD_CAST "PO_REQ",
@@ -74,13 +81,15 @@ int main(void) {
         BAD_CAST "PFO_ADVOK",
         BAD_CAST "PFO_DANPOR",
         BAD_CAST "PFO_INSSPR",
+        BAD_CAST "PFO_AUDITOR",
         BAD_CAST "FO"
     };
 
-    TEST("DBTYPE_SYSTEM", test_dbtype2string_must_fail, DBTYPE_SYSTEM); 
+    TEST("DBTYPE_OVM_MAIN", test_dbtype2string_must_fail, DBTYPE_OVM_MAIN);
+    TEST("DBTYPE_SYSTEM", test_dbtype2string_must_fail, DBTYPE_SYSTEM);
 
     for (size_t i = 0; i < sizeof(types)/sizeof(types[0]); i++)
-        TEST(isds_DbType2string(types[i]), test_dbtype, types[i], names[i]);
+        TEST(names[i], test_dbtype, types[i], names[i]);
 
     TEST("1234", test_dbtype2string_must_fail, 1234); 
     
