@@ -55,7 +55,9 @@ int main(void) {
         USERTYPE_ADMINISTRATOR,
         USERTYPE_OFFICIAL,
         USERTYPE_OFFICIAL_CERT,
-        USERTYPE_LIQUIDATOR
+        USERTYPE_LIQUIDATOR,
+        USERTYPE_RECEIVER,
+        USERTYPE_GUARDIAN
     };
 
     const xmlChar *names[] = {
@@ -64,12 +66,14 @@ int main(void) {
         BAD_CAST "ADMINISTRATOR",
         BAD_CAST "OFFICIAL",
         BAD_CAST "OFFICIAL_CERT",
-        BAD_CAST "LIQUIDATOR"
+        BAD_CAST "LIQUIDATOR",
+        BAD_CAST "RECEIVER",
+        BAD_CAST "GUARDIAN"
     };
 
 
     for (size_t i = 0; i < sizeof(types)/sizeof(types[0]); i++)
-        TEST(isds_UserType2string(types[i]), test_usertype, types[i], names[i]);
+        TEST(names[i], test_usertype, types[i], names[i]);
 
     TEST("1234", test_usertype2string_must_fail, 1234); 
     
