@@ -23,7 +23,7 @@ static char *server_certificate = TLSDIR "/server.cert";
 static char *server_key = TLSDIR "/server.key";
 static char *client_certificate = TLSDIR "/client.cert";
 static char *client_key = TLSDIR "/client.key";
-static const char *client_dn = "C=CZ,CN=The Client";
+static const char *client_dn = "CN=The Client,C=CZ";
 static const char *username = "douglas";
 static const char *password = "42";
 
@@ -35,7 +35,11 @@ static int test_login(const isds_error error, struct isds_ctx *context,
     isds_error err;
 
     err = isds_login(context, url, username, password, pki_credentials, otp);
-    if (error != err)
+    /* If TLSv1.3 is used, cURL reports a network error instead of a security
+     * error if server refuses client's certificate (since GnuTLS 3.6.4).
+     * Maybe a <https://gitlab.com/gnutls/gnutls/issues/615>. As a workaround,
+     * accept IE_NETWORK if IE_SECURITY was expected. */
+    if (error != err && (IE_SECURITY != error || IE_NETWORK != err))
         FAIL_TEST("Wrong return code: expected=%s, returned=%s (%s)",
                 isds_strerror(error), isds_strerror(err),
                 isds_long_message(context));
