@@ -6,6 +6,8 @@
 #include "system.h"
 #include <string.h>
 #include <stdint.h>     /* For intmax_t */
+#include <inttypes.h>   /* For PRIdMAX */
+#include <ctype.h>      /* for isdigit() */
 #include <libxml/parser.h>
 #include <libxml/xpath.h>
 #include <libxml/xpathInternals.h>
@@ -46,6 +48,161 @@ struct service {
             const void *arguments);
 };
 
+
+/* Convert UTF-8 ISO 8601 date-time @string to struct timeval.
+ * It respects microseconds too. Microseconds are rounded half up.
+ * In case of error, @time will be freed. */
+static http_error timestring2timeval(const char *string,
+        struct timeval **time) {
+    struct tm broken;
+    char *offset, *delim, *endptr;
+    const int subsecond_resolution = 6;
+    char subseconds[subsecond_resolution + 1];
+    _Bool round_up = 0;
+    int offset_hours, offset_minutes;
+    int i;
+    long int long_number;
+#ifdef _WIN32
+    int tmp;
+#endif
+    
+    if (!time) return HTTP_ERROR_SERVER;
+    if (!string) {
+        free(*time);
+        *time = NULL;
+        return HTTP_ERROR_CLIENT;
+    }
+
+    memset(&broken, 0, sizeof(broken));
+
+    if (!*time) {
+        *time = calloc(1, sizeof(**time));
+        if (!*time) return HTTP_ERROR_SERVER;
+    } else {
+        memset(*time, 0, sizeof(**time));
+    }
+
+
+    /* xsd:date is ISO 8601 string, thus ASCII */
+    /*TODO: negative year */
+
+#ifdef _WIN32
+    i = 0;
+    if ((tmp = sscanf((const char*)string, "%d-%d-%dT%d:%d:%d%n",
+        &broken.tm_year, &broken.tm_mon, &broken.tm_mday,
+        &broken.tm_hour, &broken.tm_min, &broken.tm_sec,
+        &i)) < 6) {
+        free(*time);
+        *time = NULL;
+        return HTTP_ERROR_CLIENT;
+    }
+
+    broken.tm_year -= 1900;
+    broken.tm_mon--;
+    broken.tm_isdst = -1;
+    offset = (char*)string + i;
+#else
+    /* Parse date and time without subseconds and offset */
+    offset = strptime((char*)string, "%Y-%m-%dT%T", &broken);
+    if (!offset) {
+        free(*time);
+        *time = NULL;
+        return HTTP_ERROR_CLIENT;
+    }
+#endif
+    
+    /* Get subseconds */
+    if (*offset == '.' ) {
+        offset++;
+
+        /* Copy first 6 digits, pad it with zeros.
+         * Current server implementation uses only millisecond resolution. */
+        /* TODO: isdigit() is locale sensitive */
+        for (i = 0;
+                i < subsecond_resolution && isdigit(*offset);
+                i++, offset++) {
+            subseconds[i] = *offset;
+        }
+        if (subsecond_resolution == i && isdigit(*offset)) {
+            /* Check 7th digit for rounding */
+            if (*offset >= '5') round_up = 1;
+            offset++;
+        }
+        for (; i < subsecond_resolution; i++) {
+            subseconds[i] = '0';
+        }
+        subseconds[subsecond_resolution] = '\0';
+
+        /* Convert it into integer */
+        long_number = strtol(subseconds, &endptr, 10);
+        if (*endptr != '\0' || long_number == LONG_MIN ||
+                long_number == LONG_MAX) {
+            free(*time);
+            *time = NULL;
+            return HTTP_ERROR_SERVER;
+        }
+        /* POSIX sys_time.h(0p) defines tv_usec timeval member as su_seconds_t
+         * type. sys_types.h(0p) defines su_seconds_t as "used for time in
+         * microseconds" and "the type shall be a signed integer capable of
+         * storing values at least in the range [-1, 1000000]. */
+        if (long_number < -1 || long_number >= 1000000) {
+            free(*time);
+            *time = NULL;
+            return HTTP_ERROR_CLIENT;
+        }
+        (*time)->tv_usec = long_number;
+
+        /* Round the subseconds */
+        if (round_up) {
+            if (999999 == (*time)->tv_usec) {
+                (*time)->tv_usec = 0;
+                broken.tm_sec++;
+            } else {
+                (*time)->tv_usec++;
+            }
+        }
+
+        /* move to the zone offset delimiter or signal NULL*/
+        delim = strchr(offset, '-');
+        if (!delim)
+            delim = strchr(offset, '+');
+        if (!delim)
+            delim = strchr(offset, 'Z');
+        offset = delim;
+    }
+
+    /* Get zone offset */
+    /* ISO allows zone offset string only: "" | "Z" | ("+"|"-" "<HH>:<MM>")
+     * "" equals to "Z" and it means UTC zone. */
+    /* One can not use strptime(, "%z",) becase it's RFC E-MAIL format without
+     * colon separator */
+    if (offset && (*offset == '-' || *offset == '+')) {
+        if (2 != sscanf(offset + 1, "%2d:%2d", &offset_hours, &offset_minutes)) {
+            free(*time);
+            *time = NULL;
+            return HTTP_ERROR_CLIENT;
+        }
+        if (*offset == '+') {
+            broken.tm_hour -= offset_hours;
+            broken.tm_min -= offset_minutes;
+        } else {
+            broken.tm_hour += offset_hours;
+            broken.tm_min += offset_minutes;
+        }
+    }
+
+    /* Convert to time_t */
+    (*time)->tv_sec = _isds_timegm(&broken);
+    if ((*time)->tv_sec == (time_t) -1) {
+        free(*time);
+        *time = NULL;
+        return HTTP_ERROR_CLIENT;
+    }
+
+    return HTTP_ERROR_SUCCESS;
+}
+
+
 /* Following EXTRACT_* macros expect @xpath_ctx, @error, @message,
  * and leave label. */
 #define ELEMENT_EXISTS(element, allow_multiple) { \
@@ -278,6 +435,42 @@ static http_error element_exists(const char **code, char **message,
 }
 
 
+/* Locate a children element.
+ * @code is a static output ISDS error code
+ * @error_message is a reallocated output ISDS error message
+ * @xpath_ctx is a current XPath context
+ * @element_name is name of an element to select
+ * @node is output pointer to located element node
+ * @return HTTP_ERROR_SUCCESS or an appropriate error code. */
+static http_error select_element(const char **code, char **message,
+        xmlXPathContextPtr xpath_ctx, const char *element_name,
+        xmlNodePtr *node) {
+    xmlXPathObjectPtr result = NULL;
+
+    result = xmlXPathEvalExpression(BAD_CAST element_name, xpath_ctx);
+    if (NULL == result) {
+        return HTTP_ERROR_SERVER;
+    }
+    if (xmlXPathNodeSetIsEmpty(result->nodesetval)) {
+            xmlXPathFreeObject(result);
+            *code = "9999";
+            test_asprintf(message, "Element %s does not exist", element_name);
+            return HTTP_ERROR_CLIENT;
+    } else {
+        if (result->nodesetval->nodeNr > 1) {
+            xmlXPathFreeObject(result);
+            *code = "9999";
+            test_asprintf(message, "Multiple %s element", element_name);
+            return HTTP_ERROR_CLIENT;
+        }
+    }
+    *node = result->nodesetval->nodeTab[0];
+    xmlXPathFreeObject(result);
+
+    return HTTP_ERROR_SUCCESS;
+}
+
+
 /* Extract @element_name's value as a string.
  * @code is a static output ISDS error code
  * @error_message is a reallocated output ISDS error message
@@ -326,6 +519,29 @@ leave:
 }
 
 
+/* Compare dates represented by pointer to struct tm.
+ * @return 0 if equalued, non-0 otherwise. */
+static int datecmp(const struct tm *a, const struct tm *b) {
+    if (NULL == a && b == NULL) return 0;
+    if ((NULL == a && b != NULL) || (NULL != a && b == NULL)) return 1;
+    if (a->tm_year != b->tm_year) return 1;
+    if (a->tm_mon != b->tm_mon) return 1;
+    if (a->tm_mday != b->tm_mday) return 1;
+    return 0;
+}
+
+
+/* Compare times represented by pointer to struct timeval.
+ * @return 0 if equalued, non-0 otherwise. */
+static int timecmp(const struct timeval *a, const struct timeval *b) {
+    if (NULL == a && b == NULL) return 0;
+    if ((NULL == a && b != NULL) || (NULL != a && b == NULL)) return 1;
+    if (a->tv_sec != b->tv_sec) return 1;
+    if (a->tv_usec != b->tv_usec) return 1;
+    return 0;
+}
+
+
 /* Checks an @element_name's value is an @expected_value string.
  * @code is a static output ISDS error code
  * @error_message is a reallocated output ISDS error message
@@ -540,6 +756,149 @@ leave:
 }
 
 
+/* Checks an @element_name's value is an @expected_value date.
+ * @code is a static output ISDS error code
+ * @error_message is an reallocated output ISDS error message
+ * @xpath_ctx is a current XPath context
+ * @element_name is name of a element to check
+ * @must_exist is true if the @element_name must exist even if @expected_value
+ * is NULL.
+ * @expected_value is an expected boolean value
+ * @return HTTP_ERROR_SUCCESS if the @element_name element's value is
+ * @expected_value. HTTP_ERROR_CLIENT if not equaled, HTTP_ERROR_SERVER if an
+ * internal error occured. */
+static http_error element_equals_date(const char **code, char **message,
+        xmlXPathContextPtr xpath_ctx, const char *element_name,
+        _Bool must_exist, const struct tm *expected_value) {
+    http_error error = HTTP_ERROR_SUCCESS;
+    char *string = NULL;
+    struct tm value;
+
+    if (must_exist) {
+        error = element_exists(code, message, xpath_ctx, element_name, 0);
+        if (HTTP_ERROR_SUCCESS != error)
+            goto leave;
+    }
+
+    error = extract_string(code, message, xpath_ctx, element_name, &string);
+    if (HTTP_ERROR_SUCCESS != error)
+        goto leave;
+
+    if (NULL != expected_value) {
+        if (NULL == string) {
+            *code = "9999";
+            test_asprintf(message, "Empty %s element", element_name);
+            error = HTTP_ERROR_CLIENT;
+            goto leave;
+        }
+        error = _server_datestring2tm(string, &value);
+        if (error) {
+            if (error == HTTP_ERROR_CLIENT) { \
+                test_asprintf(message, "%s value is not a valid date: %s",
+                        element_name, string);
+            }
+            goto leave;
+        }
+        if (datecmp(expected_value, &value)) {
+            *code = "9999";
+            test_asprintf(message, "Unexpected %s element value: "
+                    "expected=%d-%02d-%02d, got=%d-%02d-%02d", element_name,
+                    expected_value->tm_year + 1900, expected_value->tm_mon + 1,
+                    expected_value->tm_mday,
+                    value.tm_year + 1900, value.tm_mon + 1, value.tm_mday);
+            error = HTTP_ERROR_CLIENT;
+            goto leave;
+        }
+    } else {
+        if (NULL != string && *string != '\0') {
+            *code = "9999";
+            test_asprintf(message,
+                    "Unexpected %s element value: "
+                    "expected empty value, got=`%s'",
+                    element_name, string);
+            error = HTTP_ERROR_CLIENT;
+            goto leave;
+        }
+    }
+
+leave:
+    free(string);
+    return error;
+}
+
+
+/* Checks an @element_name's value is an @expected_value time.
+ * @code is a static output ISDS error code
+ * @error_message is an reallocated output ISDS error message
+ * @xpath_ctx is a current XPath context
+ * @element_name is name of a element to check
+ * @must_exist is true if the @element_name must exist even if @expected_value
+ * is NULL.
+ * @expected_value is an expected boolean value
+ * @return HTTP_ERROR_SUCCESS if the @element_name element's value is
+ * @expected_value. HTTP_ERROR_CLIENT if not equaled, HTTP_ERROR_SERVER if an
+ * internal error occured. */
+static http_error element_equals_time(const char **code, char **message,
+        xmlXPathContextPtr xpath_ctx, const char *element_name,
+        _Bool must_exist, const struct timeval *expected_value) {
+    http_error error = HTTP_ERROR_SUCCESS;
+    char *string = NULL;
+    struct timeval *value = NULL;
+
+    if (must_exist) {
+        error = element_exists(code, message, xpath_ctx, element_name, 0);
+        if (HTTP_ERROR_SUCCESS != error)
+            goto leave;
+    }
+
+    error = extract_string(code, message, xpath_ctx, element_name, &string);
+    if (HTTP_ERROR_SUCCESS != error)
+        goto leave;
+
+    if (NULL != expected_value) {
+        if (NULL == string) {
+            *code = "9999";
+            test_asprintf(message, "Empty %s element", element_name);
+            error = HTTP_ERROR_CLIENT;
+            goto leave;
+        }
+        error = timestring2timeval(string, &value);
+        if (error) {
+            if (error == HTTP_ERROR_CLIENT) { \
+                test_asprintf(message, "%s value is not a valid time: %s",
+                        element_name, string);
+            }
+            goto leave;
+        }
+        if (timecmp(expected_value, value)) {
+            *code = "9999";
+            test_asprintf(message, "Unexpected %s element value: "
+                    "expected=%ds:%" PRIdMAX "us, got=%ds:%" PRIdMAX "us",
+                    element_name,
+                    expected_value->tv_sec, (intmax_t)expected_value->tv_usec,
+                    value->tv_sec, (intmax_t)value->tv_usec);
+            error = HTTP_ERROR_CLIENT;
+            goto leave;
+        }
+    } else {
+        if (NULL != string && *string != '\0') {
+            *code = "9999";
+            test_asprintf(message,
+                    "Unexpected %s element value: "
+                    "expected empty value, got=`%s'",
+                    element_name, string);
+            error = HTTP_ERROR_CLIENT;
+            goto leave;
+        }
+    }
+
+leave:
+    free(string);
+    free(value);
+    return error;
+}
+
+
 /* Insert dmStatus or similar subtree
  * @parent is element to insert to
  * @dm is true for dmStatus, otherwise dbStatus
@@ -599,7 +958,7 @@ static http_error timeval2timestring(const struct timeval *time,
      * integer to accomodate values from range [-1, 1000000].
      * See <http://www.w3.org/TR/2001/REC-xmlschema-2-20010502/#dateTime> */ 
     if (-1 == test_asprintf(string,
-                "%04d-%02d-%02dT%02d:%02d:%02d.%06jd",
+                "%04d-%02d-%02dT%02d:%02d:%02d.%06" PRIdMAX,
                 broken.tm_year + 1900, broken.tm_mon + 1, broken.tm_mday,
                 broken.tm_hour, broken.tm_min, broken.tm_sec,
                 (intmax_t)time->tv_usec))
@@ -609,18 +968,6 @@ static http_error timeval2timestring(const struct timeval *time,
 }
 
 
-/* Compare dates represented by pointer to struct tm.
- * @return 0 if equalued, non-0 otherwise. */
-static int datecmp(const struct tm *a, const struct tm *b) {
-    if (NULL == a && b == NULL) return 0;
-    if ((NULL == a && b != NULL) || (NULL != a && b == NULL)) return 1;
-    if (a->tm_year != b->tm_year) return 1;
-    if (a->tm_mon != b->tm_mon) return 1;
-    if (a->tm_mday != b->tm_mday) return 1;
-    return 0;
-}
-
-
 /* Implement DummyOperation */
 static http_error service_DummyOperation(
         xmlXPathContextPtr xpath_ctx,
@@ -943,6 +1290,310 @@ leave:
 }
 
 
+/* Insert list of search results as XSD:tDbOwnersArray XML tree.
+ * @isds_response is XML node with the response
+ * @results is list of struct server_owner_info *.
+ * @create_empty_root is true to create dbResults element even if @results is
+ * empty. */
+static http_error insert_tDbOwnersArray(xmlNodePtr isds_response,
+        const struct server_list *results, _Bool create_empty_root) {
+    http_error error = HTTP_ERROR_SUCCESS;
+    xmlNodePtr root, entry;
+
+    if (NULL == isds_response) return HTTP_ERROR_SERVER;
+
+    if (NULL != results || create_empty_root)
+        INSERT_ELEMENT(root, isds_response, "dbResults");
+
+    if (NULL == results) return HTTP_ERROR_SUCCESS;
+
+    for (const struct server_list *item = results; NULL != item;
+            item = item->next) {
+        const struct server_owner_info *result =
+            (struct server_owner_info *)item->data;
+        
+        INSERT_ELEMENT(entry, root, "dbOwnerInfo");
+        if (NULL == result) continue;
+
+        INSERT_STRING(entry, "dbID", result->dbID);
+        /* Ignore aifoIsds */
+        INSERT_STRING(entry, "dbType", result->dbType);
+        INSERT_STRING(entry, "ic", result->ic);
+        INSERT_STRING(entry, "pnFirstName", result->pnFirstName);
+        INSERT_STRING(entry, "pnMiddleName", result->pnMiddleName);
+        INSERT_STRING(entry, "pnLastName", result->pnLastName);
+        INSERT_STRING(entry, "pnLastNameAtBirth", result->pnLastNameAtBirth);
+        INSERT_STRING(entry, "firmName", result->firmName);
+        INSERT_TMPTR(entry, "biDate", result->biDate);
+        INSERT_STRING(entry, "biCity", result->biCity);
+        INSERT_STRING(entry, "biCounty", result->biCounty);
+        INSERT_STRING(entry, "biState", result->biState);
+        /* Ignore adCode */
+        INSERT_STRING(entry, "adCity", result->adCity);
+        /* Ignore adDistrict */
+        INSERT_STRING(entry, "adStreet", result->adStreet);
+        INSERT_STRING(entry, "adNumberInStreet", result->adNumberInStreet);
+        INSERT_STRING(entry, "adNumberInMunicipality",
+                result->adNumberInMunicipality);
+        INSERT_STRING(entry, "adZipCode", result->adZipCode);
+        INSERT_STRING(entry, "adState", result->adState);
+        INSERT_STRING(entry, "nationality", result->nationality);
+        if (result->email_exists || result->email != NULL) {
+            INSERT_STRING(entry, "email", result->email);
+        }
+        if (result->telNumber_exists || result->telNumber != NULL) {
+            INSERT_STRING(entry, "telNumber", result->telNumber);
+        }
+        INSERT_STRING(entry, "identifier", result->identifier);
+        INSERT_STRING(entry, "registryCode", result->registryCode);
+        INSERT_LONGINTPTR(entry, "dbState", result->dbState);
+        INSERT_BOOLEANPTR(entry, "dbEffectiveOVM", result->dbEffectiveOVM);
+        INSERT_BOOLEANPTR(entry, "dbOpenAddressing", result->dbOpenAddressing);
+    }
+
+leave:
+    return error;
+}
+
+
+/* Implement FindDataBox.
+ * @arguments is pointer to struct arguments_DS_df_FindDataBox */
+static http_error service_FindDataBox(
+        xmlXPathContextPtr xpath_ctx,
+        xmlNodePtr isds_response,
+        const void *arguments) {
+    http_error error = HTTP_ERROR_SUCCESS;
+    const char *code = "9999";
+    char *message = NULL;
+    const struct arguments_DS_df_FindDataBox *configuration =
+        (const struct arguments_DS_df_FindDataBox *)arguments;
+    char *string = NULL;
+    xmlNodePtr node;
+
+    if (NULL == configuration || NULL == configuration->status_code ||
+            NULL == configuration->status_message) {
+        error = HTTP_ERROR_SERVER;
+        goto leave;
+    }
+
+    /* Check request */
+    error = select_element(&code, &message, xpath_ctx, "isds:dbOwnerInfo",
+            &node);
+    if (error) goto leave;
+    xpath_ctx->node = node;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:dbID", 1, configuration->criteria->dbID);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:dbType", 1, configuration->criteria->dbType);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:ic", 1, configuration->criteria->ic);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:pnFirstName", 1, configuration->criteria->pnFirstName);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:pnMiddleName", 1, configuration->criteria->pnMiddleName);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:pnLastName", 1, configuration->criteria->pnLastName);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:pnLastNameAtBirth", 1, configuration->criteria->pnLastNameAtBirth);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:firmName", 1, configuration->criteria->firmName);
+    if (error) goto leave;
+
+    error = element_equals_date(&code, &message, xpath_ctx,
+            "isds:biDate", 1, configuration->criteria->biDate);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:biCity", 1, configuration->criteria->biCity);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:biCounty", 1, configuration->criteria->biCounty);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:biState", 1, configuration->criteria->biState);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:adCity", 1, configuration->criteria->adCity);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:adStreet", 1, configuration->criteria->adStreet);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:adNumberInStreet", 1,
+            configuration->criteria->adNumberInStreet);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:adNumberInMunicipality", 1,
+            configuration->criteria->adNumberInMunicipality);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:adZipCode", 1, configuration->criteria->adZipCode);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:adState", 1, configuration->criteria->adState);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:nationality", 1, configuration->criteria->nationality);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:email", 0, configuration->criteria->email);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:telNumber", 0, configuration->criteria->telNumber);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:identifier", 1, configuration->criteria->identifier);
+    if (error) goto leave;
+
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:registryCode", 1, configuration->criteria->registryCode);
+    if (error) goto leave;
+
+    error = element_equals_integer(&code, &message, xpath_ctx,
+            "isds:dbState", 1, configuration->criteria->dbState);
+    if (error) goto leave;
+
+    error = element_equals_boolean(&code, &message, xpath_ctx,
+            "isds:dbEffectiveOVM", 1, configuration->criteria->dbEffectiveOVM);
+    if (error) goto leave;
+
+    error = element_equals_boolean(&code, &message, xpath_ctx,
+            "isds:dbOpenAddressing", 1,
+            configuration->criteria->dbOpenAddressing);
+    if (error) goto leave;
+
+    /* Build response */
+    if ((error = insert_tDbOwnersArray(isds_response, configuration->results,
+                    configuration->results_exists))) {
+        goto leave;
+    }
+
+    code = configuration->status_code;
+    message = strdup(configuration->status_message);
+
+leave:
+    if (HTTP_ERROR_SERVER != error) {
+        http_error next_error = insert_isds_status(isds_response, 0,
+                BAD_CAST code, BAD_CAST message, NULL);
+        if (HTTP_ERROR_SUCCESS != next_error) error = next_error;
+    }
+    free(string);
+    free(message);
+    return error;
+}
+
+
+/* Insert list of period results as XSD:tdbPeriodsArray XML tree.
+ * @isds_response is XML node with the response
+ * @results is list of struct server_box_state_period *.
+ * @create_empty_root is true to create Periods element even if @results is
+ * empty. */
+static http_error insert_tdbPeriodsArray(xmlNodePtr isds_response,
+        const struct server_list *results, _Bool create_empty_root) {
+    http_error error = HTTP_ERROR_SUCCESS;
+    xmlNodePtr root, entry;
+
+    if (NULL == isds_response) return HTTP_ERROR_SERVER;
+
+    if (NULL != results || create_empty_root)
+        INSERT_ELEMENT(root, isds_response, "Periods");
+
+    if (NULL == results) return HTTP_ERROR_SUCCESS;
+
+    for (const struct server_list *item = results; NULL != item;
+            item = item->next) {
+        const struct server_box_state_period *result =
+            (struct server_box_state_period *)item->data;
+        
+        INSERT_ELEMENT(entry, root, "Period");
+        if (NULL == result) continue;
+
+        INSERT_TIMEVALPTR(entry, "PeriodFrom", result->from);
+        INSERT_TIMEVALPTR(entry, "PeriodTo", result->to);
+        INSERT_LONGINTPTR(entry, "DbState", &(result->dbState));
+    }
+
+leave:
+    return error;
+}
+
+
+/* Implement GetDataBoxActivityStatus.
+ * @arguments is pointer to struct arguments_DS_df_GetDataBoxActivityStatus */
+static http_error service_GetDataBoxActivityStatus(
+        xmlXPathContextPtr xpath_ctx,
+        xmlNodePtr isds_response,
+        const void *arguments) {
+    http_error error = HTTP_ERROR_SUCCESS;
+    const char *code = "9999";
+    char *message = NULL;
+    const struct arguments_DS_df_GetDataBoxActivityStatus *configuration =
+        (const struct arguments_DS_df_GetDataBoxActivityStatus *)arguments;
+
+    if (NULL == configuration || NULL == configuration->status_code ||
+            NULL == configuration->status_message) {
+        error = HTTP_ERROR_SERVER;
+        goto leave;
+    }
+
+    /* Check request */
+    error = element_equals_string(&code, &message, xpath_ctx,
+            "isds:dbID", 1, configuration->box_id);
+    /* ??? XML schema and textual documentation does not agree on obligatority
+     * of the isds:baFrom and isds:baTo value or presence. */
+    error = element_equals_time(&code, &message, xpath_ctx,
+            "isds:baFrom", 1, configuration->from);
+    error = element_equals_time(&code, &message, xpath_ctx,
+            "isds:baTo", 1, configuration->to);
+    if (error) goto leave;
+
+    /* Build response */
+    if ((error = insert_tdbPeriodsArray(isds_response, configuration->results,
+                    configuration->results_exists))) {
+        goto leave;
+    }
+
+    code = configuration->status_code;
+    message = strdup(configuration->status_message);
+
+leave:
+    if (HTTP_ERROR_SERVER != error) {
+        http_error next_error = insert_isds_status(isds_response, 0,
+                BAD_CAST code, BAD_CAST message, NULL);
+        if (HTTP_ERROR_SUCCESS != next_error) error = next_error;
+    }
+    free(message);
+    return error;
+}
+
+
 /* Implement ISDSSearch2.
  * @arguments is pointer to struct arguments_DS_df_ISDSSearch2 */
 static http_error service_ISDSSearch2(
@@ -1283,6 +1934,12 @@ static struct service services[] = {
     { SERVICE_DS_df_DataBoxCreditInfo,
         "DS/df", BAD_CAST ISDS_NS, BAD_CAST "DataBoxCreditInfo",
         service_DataBoxCreditInfo },
+    { SERVICE_DS_df_FindDataBox,
+        "DS/df", BAD_CAST ISDS_NS, BAD_CAST "FindDataBox",
+        service_FindDataBox },
+    { SERVICE_DS_df_GetDataBoxActivityStatus,
+        "DS/df", BAD_CAST ISDS_NS, BAD_CAST "GetDataBoxActivityStatus",
+        service_GetDataBoxActivityStatus },
     { SERVICE_DS_df_ISDSSearch2,
         "DS/df", BAD_CAST ISDS_NS, BAD_CAST "ISDSSearch2",
         service_ISDSSearch2 },
