@@ -8,27 +8,30 @@
 #include <stdio.h>
 #include <locale.h>
 #include <stdint.h>
+#include <inttypes.h>
 #include <stdarg.h>
 #include <string.h>
 
 #include "test-tools.h"
 
-char *unit_name, *reason = NULL;
-unsigned int passed, failed;
-void (*test_destructor_function)(void *) = NULL;
-void *test_destructor_argument;
+extern char *unit_name;
+extern char *reason;
+extern unsigned int passed, failed, skipped;
+extern void (*test_destructor_function)(void *);
+extern void *test_destructor_argument;
 
 #define INIT_TEST(name) { \
     setlocale(LC_ALL, "C"); \
     unit_name = name; \
-    passed = failed = 0; \
+    passed = failed = skipped = 0; \
     printf("Testing unit: %s\n", unit_name); \
 }
 
 #define SUM_TEST() { \
-    printf("Test results: unit = %s, passed = %u, failed = %u\n\n", \
-            unit_name, passed, failed); \
-    exit(failed ? EXIT_FAILURE : EXIT_SUCCESS ); \
+    printf("Test results: unit = %s, passed = %u, failed = %u, skipped = %u\n\n", \
+            unit_name, passed, failed, skipped); \
+    exit(failed ? EXIT_FAILURE : passed ? EXIT_SUCCESS : \
+            skipped ? 77 : EXIT_SUCCESS); \
 }
 
 #define TEST_DESTRUCTOR(function, argument) { \
@@ -39,6 +42,7 @@ void *test_destructor_argument;
 #define PASS_TEST { \
     if (NULL != test_destructor_function) \
         test_destructor_function(test_destructor_argument); \
+    TEST_DESTRUCTOR(NULL, NULL); \
     return 0; \
 } \
 
@@ -51,9 +55,27 @@ void *test_destructor_argument;
     FAILURE_REASON(__VA_ARGS__); \
     if (NULL != test_destructor_function) \
         test_destructor_function(test_destructor_argument); \
+    TEST_DESTRUCTOR(NULL, NULL); \
     return 1; \
 }
 
+#define SKIP_TEST(...) { \
+    FAILURE_REASON(__VA_ARGS__); \
+    if (NULL != test_destructor_function) \
+        test_destructor_function(test_destructor_argument); \
+    TEST_DESTRUCTOR(NULL, NULL); \
+    return 77; \
+}
+
+#define SKIP_TESTS(number, reason) { \
+    if (1 == (number)) { \
+        printf("\t%u test skipped: %s\n", (number), (reason)); \
+    } else {\
+        printf("\t%u tests skipped: %s\n", (number), (reason)); \
+    } \
+    skipped += (number); \
+}
+
 #define ABORT_UNIT(message) { \
     printf("Unit %s procedure aborted: %s\n", unit_name, (message)); \
     exit(EXIT_FAILURE); \
@@ -65,7 +87,10 @@ void *test_destructor_argument;
     free(reason); reason = NULL; \
     test_destructor_function = NULL; \
     int status = (function)(__VA_ARGS__); \
-    if (status) { \
+    if (77 == status) { \
+        skipped++; \
+        test_message = "skipped"; \
+    } else if (status) { \
         failed++; \
         test_message = "failed"; \
     } else { \
@@ -155,7 +180,7 @@ void *test_destructor_argument;
                     "expected=%d, got=%d", (x)->tv_sec, (y)->tv_sec); \
         if ((x)->tv_usec != (y)->tv_usec) \
             FAIL_TEST("struct timeval * differs in tv_usec: " \
-                    "expected=%jd, got=%jd", (intmax_t)(x)->tv_usec, \
+                    "expected=%" PRIdMAX ", got=%" PRIdMAX, (intmax_t)(x)->tv_usec, \
                     (intmax_t)(y)->tv_usec); \
     } \
 }
